import collections.abc


def snake_to_camel(word, upper):
    if word == None:
        return None

    if word.find("_") == -1:
        return word

    parts = word.split("_")

    if upper:
        res = parts[0][0].upper() + parts[0][1:].lower()
    else:
        res = parts[0].lower()
    for i in range(1, len(parts)):
        res += parts[i][0].upper() + parts[i][1:].lower()

    return res


def dict_update(d, u):
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = dict_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


def file_write(file, data):
    if file != None:
        file.write(data + "\n")


SEPARATORS = [" ", "\n", "\r", "{", "}", "=", ">",
              "<", ":", "?", ",", "\"", "\'", "(", ")"]


def next_separator_without_spaces_index(data):
    """Get index of next separator
    that is not a space"""

    tmp_seps = SEPARATORS[:]
    tmp_seps.remove(" ")
    tmp_seps.remove("\n")
    tmp_seps.remove("\r")

    for i in range(0, len(data)):
        if data[i] in tmp_seps:
            return i

        elif data[i] in [" ", "\n", "\r"]:
            continue

        break

    return None  # EOF or non-separator


def get_next_separator_without_spaces(data):
    """Get next separator that is not a space"""

    index = next_separator_without_spaces_index(data)

    if index == None:
        return None

    return data[index]


def get_next_word_indices(data, ignore_sep):
    """Get indices between separators,
    skipping separators at start"""

    start = None

    if len(data) == 1:
        return (None, None)

    for i in range(0, len(data)):
        if start == None and not ignore_sep and data[i] in SEPARATORS:
            break

        if start == None and data[i] not in SEPARATORS:
            start = i

        if start != None and data[i] in SEPARATORS:
            return start, i

        if start != None and i == len(data) - 1:
            return start, i+1

    return (None, None)  # file read until EOF


def read_next_word(data, ignore_sep=True):
    """Read the word, disregard indices"""
    start, end = get_next_word_indices(data, ignore_sep)

    if start == None:
        return None

    return data[start:end]


def check_word_with_space(data, word):
    """Checks for a type"""

    word_len = len(word)

    if len(data) < word_len + 1:
        return False

    if data.startswith(word + " "):
        return True

    return False


def get_symbols_in_parentheses(data):
    """Get parameter names in parentheses"""

    i = 0
    res = {}

    while get_next_separator_without_spaces(data[i:]) != ")":
        word1 = read_next_word(data[i:])
        i += len(word1) + data[i:].index(word1)

        word2 = None

        if get_next_separator_without_spaces(data[i]) == None:
            word2 = read_next_word(data[i:])
            i += len(word2) + data[i:].index(word2)

        # res[outerName] = innerName
        if word2 != None:
            res[word1] = word2
        else:
            res[word1] = word1

        # disregard parameter type
        tmp = read_next_word(data[i:])
        i += len(tmp) + data[i:].index(tmp)

    return res, i


def is_next_newline(data):
    """While skipping spaces, checks for \n as next char"""
    for i in range(len(data)):
        if data[i] == "\n":
            return True

        elif data[i] in [" ", "\r"]:
            continue

        return False


def get_vars_directly_assigned(data, symbols):
    """Looking for expression "a = b" in init"""

    res = {}
    newline = False
    i = 0
    nesting = 0

    while i < len(data):
        if data[i] == "\n":
            newline = True
            i += 1
            continue

        if data[i] == "{":
            nesting += 1

        if data[i] == "}":
            nesting -= 1

            if nesting == 0:
                break

        if newline == True:
            word1 = read_next_word(data[i:])

            if word1 == None:
                break

            i += len(word1) + data[i:].index(word1)

            if get_next_separator_without_spaces(data[i:]) == "=":
                word2 = read_next_word(data[i:])
                i += len(word2) + data[i:].index(word2)

                if is_next_newline(data[i:]) and word2 in symbols.values() and not word1.startswith("self."):
                    res[word1] = word2  # word1 = word2

            newline = False

        else:
            i += 1

    return res


def apply_init_changes(file_data, symbols_in_parentheses, vars_directly_assigned, filepath):
    """Change "a = b" in init to "self.a = a" """
    i = 0
    start = None
    changes = []

    while i < len(file_data):
        word = read_next_word(file_data[i:])

        if word == None:
            break

        i += len(word) + file_data[i:].index(word)

        if word == "init":
            start = i

    if start == None:
        return None

    unprocessed = file_data[start:]

    new_file_data = file_data[:start]

    # modify parameters to match type fields
    for key in symbols_in_parentheses.keys():
        if symbols_in_parentheses[key] in vars_directly_assigned.values():
            start_change = unprocessed.index(key)
            new_file_data += unprocessed[:start_change]
            unprocessed = unprocessed[start_change:]

            word1 = read_next_word(unprocessed)
            offset = len(word1) + unprocessed.index(word1)
            unprocessed = unprocessed[offset:]

            word2 = read_next_word(unprocessed)
            offset = len(word2) + unprocessed.index(word2)
            unprocessed = unprocessed[offset:]

            for k, v in vars_directly_assigned.items():
                if symbols_in_parentheses[key] == v:
                    new_file_data += k
                    break

    init_start = unprocessed.index("{")
    new_file_data += unprocessed[:init_start]
    unprocessed = unprocessed[init_start:]

    for key in vars_directly_assigned.keys():
        start_change = unprocessed.index(key)

        new_file_data += unprocessed[:start_change]
        unprocessed = unprocessed[start_change:]

        word1 = read_next_word(unprocessed)
        offset = len(word1) + unprocessed.index(word1)
        unprocessed = unprocessed[offset:]

        word2 = read_next_word(unprocessed)
        offset = len(word2) + unprocessed.index(word2)
        unprocessed = unprocessed[offset:]

        new_file_data += f"self.{key} = {key}"

        changes.append(
            f"{filepath} Changed: parameter in init was changed to class field {key}")

    new_file_data += unprocessed

    return new_file_data, changes


def apply_static_class_prop_changes(data, all_changes, filepath):
    """Apply removing suffix from static and class properties"""
    nesting = 0
    new_file_data = ""
    changes = []
    prop_changes = None

    while len(data) > 0:
        for class_change in all_changes.keys():
            if data.startswith("class " + class_change):
                prop_changes = all_changes[class_change]

        if prop_changes != None:
            if data[0] == "{":
                nesting += 1

            if data[0] == "}":
                nesting -= 1

                if nesting == 0:
                    prop_changes = None
                    continue

            for change_prop in prop_changes.keys():
                if data.startswith(change_prop):
                    new_file_data += prop_changes[change_prop]
                    data = data[len(change_prop):]
                    changes.append(
                        f"{filepath} Changed: class/static property {change_prop} (doesn't need return suffix)")
                    break

        new_file_data += data[:1]
        data = data[1:]

    return new_file_data, changes


def read_until_newline(data):
    """Read data until newline is found"""
    res = ""
    if data[0] == "\n":
        data = data[1:]

    while len(data) > 0:
        if data[0] != "\n":
            res += data[:1]
            data = data[1:]

        else:
            return res
    return res


def remove_star_in_line(line):
    """Return (docstring) line without javadoc style *"""
    for i in range(0, len(line)):
        if line[i] == " ":
            continue

        elif line[i:i+2] == "*/":
            return None

        elif line[i:i+2] == "* ":
            return line[i+2:]

        elif line[i] == "*":
            return line[i+1:]

        else:
            return line

    return ""


def process_doc_block(data):
    """Gather info about current doc block"""
    i = 0
    comment_lines = []
    flag = 1
    flags = {"Parameter": 0, "Parameters": 0,
             "Returns": 0, "Throws": 0}

    count = {"Parameter": 0, "Parameters": 0,
             "Returns": 0, "Throws": 0, "Params amount": 0}

    while i < len(data):
        line = read_until_newline(data[i:])

        if not line.lstrip().startswith("///"):
            return comment_lines, flags, count, i-1

        comment_lines.append(line)
        i += len(line) + 1  # +1 accounts for skipped \n

        try:
            line.index("- Parameters")
            flags["Parameters"] = flag
            count["Parameters"] += 1
            flag += 1
        except ValueError:

            try:
                line.index("- Parameter")
                flags["Parameter"] = flag
                count["Parameter"] += 1
                flag += 1
            except ValueError:
                pass

        try:
            line.index("- Returns")
            flags["Returns"] = flag
            count["Returns"] += 1
            flag += 1
        except ValueError:
            pass

        try:
            line.index("- Throws")
            flags["Throws"] = flag
            count["Throws"] += 1
            flag += 1
        except ValueError:
            pass

        try:
            line.index("///   - ")
            count["Params amount"] += 1
        except ValueError:
            pass


def check_tags_order(flags):
    """Check if tags go in allowed order"""
    if flags["Parameter"] == 1 and flags["Parameter"] < flags["Returns"] < flags["Throws"] and flags["Parameters"] == 0:
        return True

    if flags["Parameters"] == 1 and flags["Parameters"] < flags["Returns"] < flags["Throws"] and flags["Parameter"] == 0:
        return True

    if flags["Parameter"] == 1 and flags["Parameter"] < flags["Returns"] and flags["Throws"] == 0 and flags["Parameters"] == 0:
        return True

    if flags["Parameters"] == 1 and flags["Parameters"] < flags["Returns"] and flags["Throws"] == 0 and flags["Parameter"] == 0:
        return True

    if flags["Parameter"] == 1 and flags["Parameter"] < flags["Throws"] and flags["Returns"] == 0 and flags["Parameters"] == 0:
        return True

    if flags["Parameters"] == 1 and flags["Parameters"] < flags["Throws"] and flags["Returns"] == 0 and flags["Parameter"] == 0:
        return True

    if flags["Parameter"] == 1 and flags["Returns"] == 0 and flags["Throws"] == 0 and flags["Parameters"] == 0:
        return True

    if flags["Parameters"] == 1 and flags["Returns"] == 0 and flags["Throws"] == 0 and flags["Parameter"] == 0:
        return True

    return False


def redo_comment_block(comment_lines, count, filename):
    """Modify comment block, so that if follows code conventions"""
    new_comment_lines = []
    last_non_tag = 0
    warnings = []
    changes = []

    for line in comment_lines:
        if not line[3:].lstrip().startswith("- "):
            new_comment_lines.append(line)
            last_non_tag += 1
            continue

        break

    tags_start = last_non_tag

    if count["Parameter"] == 1:
        for i in range(tags_start, len(comment_lines)):
            try:
                comment_lines[i].index("- Parameter")
                new_comment_lines.append(comment_lines[i])

                # check next ones continuing current
                for j in range(i+1, len(comment_lines)):
                    if not comment_lines[j].startswith("/// - ") and not comment_lines[j].startswith("///   - "):
                        new_comment_lines.append(comment_lines[j])

                    else:
                        break

                break
            except ValueError:
                pass

    elif count["Parameter"] > 1:
        warnings.append(
            f"{filename} Error: When there are more than 2 parameters use \"Parameters\" instead of \"Parameter\"")

        changes.append(
            f"{filename} Changed: multiple \"Parameter\" changed to \"Parameters\"")

        new_comment_lines.append("/// - Parameters:")

        for i in range(tags_start, len(comment_lines)):
            try:
                param_index = comment_lines[i].index("- Parameter")
                new_comment_lines.append(
                    "///   -" + comment_lines[i][param_index + 11:])

                # check next ones continuing current
                for j in range(i+1, len(comment_lines)):
                    if not comment_lines[j].startswith("/// - ") and not comment_lines[j].startswith("///   - "):
                        new_comment_lines.append(comment_lines[j])

                    else:
                        break

            except ValueError:
                pass

    if count["Parameters"] == 1:
        if count["Params amount"] == 1:
            warnings.append(
                f"{filename} Error: When there is 1 parameter use \"Parameter\" instead of \"Parameters\"")

            changes.append(
                f"{filename} Changed: single parameter in \"Parameters\" changed to \"Parameter\"")

            for i in range(tags_start, len(comment_lines)):
                try:
                    param_index = comment_lines[i].index("///   - ")
                    new_comment_lines.append(
                        "/// - Parameter " + comment_lines[i][param_index + 8:])

                    # check next ones continuing current
                    for j in range(i+1, len(comment_lines)):
                        if not comment_lines[j].startswith("/// - ") and not comment_lines[j].startswith("///   - "):
                            new_comment_lines.append(comment_lines[j])

                        else:
                            break

                    break
                except ValueError:
                    pass

        else:
            new_comment_lines.append("/// - Parameters:")

            for i in range(tags_start, len(comment_lines)):
                try:
                    param_index = comment_lines[i].index("///   - ")
                    new_comment_lines.append(comment_lines[i])

                    # check next ones continuing current
                    for j in range(i+1, len(comment_lines)):
                        if not comment_lines[j].startswith("/// - ") and not comment_lines[j].startswith("///   - "):
                            new_comment_lines.append(comment_lines[j])

                        else:
                            break

                except ValueError:
                    pass

    if count["Returns"] == 1:
        for i in range(tags_start, len(comment_lines)):
            try:
                param_index = comment_lines[i].index("- Returns")
                new_comment_lines.append(comment_lines[i])

                # check next ones continuing current
                for j in range(i+1, len(comment_lines)):
                    if not comment_lines[j].startswith("/// - ") and not comment_lines[j].startswith("///   - "):
                        new_comment_lines.append(comment_lines[j])

                    else:
                        break

                break

            except ValueError:
                pass

    if count["Throws"] == 1:
        for i in range(tags_start, len(comment_lines)):
            try:
                param_index = comment_lines[i].index("- Throws")
                new_comment_lines.append(comment_lines[i])

                # check next ones continuing current
                for j in range(i+1, len(comment_lines)):
                    if not comment_lines[j].startswith("/// - ") and not comment_lines[j].startswith("///   - "):
                        new_comment_lines.append(comment_lines[j])

                    else:
                        break

                break

            except ValueError:
                pass

    return new_comment_lines, changes, warnings


def lower_camel_case(symbol):
    return snake_to_camel(symbol[0].lower() + symbol[1:], False)


def upper_camel_case(symbol):
    return snake_to_camel(symbol[0].upper() + symbol[1:], True)



import argparse
from pathlib import Path

from analyzers import *

# folder to store results
results_folder = Path("results/")

# specify file extention of interest
extention = "swift"

# create  and configure argparse
arg_parser = argparse.ArgumentParser(description="Swift CCF")

group1 = arg_parser.add_mutually_exclusive_group(required=True)
group1.add_argument("-d", action="store", dest="directory",
                    help="directory to parse")
group1.add_argument("-f", action="store", dest="file", help="file to parse")
group1.add_argument("-p", action="store", dest="project",
                    help="project directory to parse (recursive)")

arg_parser.add_argument("-v", "--verify", action="store_true",
                        dest="verify", help="output unfixable errors and warnings")
arg_parser.add_argument("-fx", "--fix", action="store_true", dest="fix",
                        help="output fixed errors")

args = arg_parser.parse_args()


def scan_and_fix_file(filepath, outname):
    """Apply usual fixers"""
    data = ""
    verify = None
    fixing = None

    with open(filepath, "r") as infile:
        data = infile.read()
        new_filename = None

        results_folder.mkdir(exist_ok=True)

        if args.verify:
            verify = open(results_folder /
                          (outname + "_verification.log"), "a+")

        if args.fix:
            fixing = open(results_folder / (outname + "_fixing.log"), "a+")

        # exit without doing anything if both none
        if verify == None and fixing == None:
            return

        for fixer in source_fixers:
            possible_new_filename = fixer(
                verify, fixing, filepath, data, filepath.stem)
            if possible_new_filename != None:
                new_filename = possible_new_filename

        for fixer in naming_fixers:
            changed_data = fixer(verify, fixing, filepath, data)
            if changed_data != None:
                data = changed_data

        for fixer in docs_fixers:
            changed_data = fixer(verify, fixing, filepath, data)
            if changed_data != None:
                data = changed_data

        if verify != None:
            verify.close()

        if fixing != None:
            fixing.close()

    if fixing != None:
        if new_filename != None:
            filepath.unlink()
            filepath = filepath.parent / (new_filename + "." + extention)
        with open(filepath, "w") as outfile:
            outfile.write(data)


def final_fix(filepath):
    """Apply global changes"""
    data = ""

    with open(filepath, "r") as infile:
        data = infile.read()

        data = apply_global(data)

    if args.fix:
        with open(filepath, "w") as outfile:
            outfile.write(data)


def main():
    if args.file:
        file = Path(args.file)

        outname = file.stem

        if not file.exists():
            raise FileNotFoundError("File not found")

        scan_and_fix_file(file, outname)

        final_fix(file)

    if args.directory:
        directory = Path(args.directory)

        outname = directory.stem

        if not directory.exists():
            raise FileNotFoundError("Directory not found")

        for filepath in directory.glob("*." + extention):
            scan_and_fix_file(filepath, outname)

        for filepath in directory.glob("*." + extention):
            final_fix(filepath)

    if args.project:
        project = Path(args.project)

        outname = project.stem

        if not project.exists():
            raise FileNotFoundError("Project directory not found")

        for filepath in project.glob("**/*." + extention):
            scan_and_fix_file(filepath, outname)

        for filepath in project.glob("**/*." + extention):
            final_fix(filepath)


if __name__ == "__main__":
    main()



from util.util import *
import string

global_changes = {}


def apply_global(file_data):
    """Applies changes implemented in
    static and class property
    and camel case changes globally"""
    i = 0
    # watch_list[type] = relative nesting scope, paretn class
    watch_list = {}
    new_file_data = ""
    last_carry_index = 0

    if global_changes.get(None):
        for variable in global_changes[None]:
            watch_list[variable] = {"nestedness": 0, "parent class": None}

    while i < len(file_data) - 1:
        word = file_data[i:i+3]

        if word in ["var", "let"]:
            i += 3

            tmp_i = i

            variable = read_next_word(file_data[tmp_i:])

            tmp_i += len(variable) + file_data[tmp_i:].index(variable)

            var_type = read_next_word(file_data[tmp_i:])

            for key in global_changes.keys():
                if key == var_type:
                    watch_list[variable] = {
                        "nestedness": 0, "parent class": var_type}

        to_pop = []

        if i < len(file_data) - 1:
            for variable in watch_list.keys():
                if file_data[i] == "{":
                    watch_list[variable]["nestedness"] += 1

                if file_data[i] == "}":
                    watch_list[variable]["nestedness"] -= 1

                if watch_list[variable]["nestedness"] < 0:
                    to_pop.append(variable)

        for p in to_pop:
            watch_list.pop(p)

        possible_change = read_next_word(file_data[i:], ignore_sep=False)

        if possible_change != None:
            for variable in watch_list.keys():
                if possible_change.startswith(variable):
                    par_class = watch_list[variable]["parent class"]
                    class_change = global_changes[par_class]

                    for change_from in class_change.keys():
                        if (possible_change.startswith(variable + "." + change_from)
                                or (par_class == None and possible_change == change_from)):

                            i = file_data.index(possible_change, i)
                            new_file_data += file_data[last_carry_index:i]
                            if par_class == None:
                                new_file_data += class_change[change_from]
                                last_carry_index = i + len(change_from)
                            else:
                                new_file_data += variable + "." + \
                                    class_change[change_from]
                                last_carry_index = i + \
                                    len(variable + "." + change_from)

        i += 1

    new_file_data += file_data[last_carry_index:]

    return new_file_data


def check_initializers(verif_file, fix_file, filepath, file_data):
    """Changes expressions of type "a=b"
    to "self.a = a" in init"""
    i = 0

    while i < len(file_data) - 1:
        word = read_next_word(file_data[i:])
        if word == None:
            return None

        i += len(word) + file_data[i:].index(word)

        if word == "init":
            symbols_in_parentheses, offset = get_symbols_in_parentheses(
                file_data[i:])
            i += offset

            vars_directly_assigned = get_vars_directly_assigned(
                file_data, symbols_in_parentheses)
            new_data, changes = apply_init_changes(
                file_data, symbols_in_parentheses, vars_directly_assigned, filepath)

            if len(changes) > 0:
                file_write(verif_file,
                           f"{filepath} Warning: init parameters direct assignment should have same name as field")

                for change in changes:
                    file_write(fix_file, change)

            return new_data


def check_properties(verif_file, fix_file, filepath, file_data):
    """Removes return suffix from class and static properties"""
    global global_changes

    i = 0
    inside_class = None
    all_changes = {}
    prop_changes = {}

    nesting = None

    while i < len(file_data):
        if check_word_with_space(file_data[i:], "class") and not check_word_with_space(file_data[i:], "class var"):
            inside_class = read_next_word(file_data[i+6:])
            prop_changes = {}
            nesting = 0

        if inside_class:
            prop = None
            prop_type = None

            if check_word_with_space(file_data[i:], "static var"):
                prop = read_next_word(file_data[i+11:])

                prop_type = read_next_word(file_data[i+11+len(prop):])

            elif check_word_with_space(file_data[i:], "class var"):
                prop = read_next_word(file_data[i+10:])

                prop_type = read_next_word(file_data[i+10+len(prop):])

            if prop != None:
                cut = -1

                for j in range(1, len(prop)):
                    if not prop_type.endswith(prop[-j:]):
                        break

                    if prop[-j] in string.ascii_uppercase:
                        cut = len(prop) - j

                if cut != -1:
                    prop_changes[prop] = prop[:cut]

            if file_data[i] == "{":
                nesting += 1

            if file_data[i] == "}":
                nesting -= 1

                if nesting == 0:  # class ended
                    all_changes[inside_class] = prop_changes

                    inside_class = None

        i += 1

    new_data, changes = apply_static_class_prop_changes(
        file_data, all_changes, filepath)

    if len(changes) > 0:
        file_write(verif_file,
                   f"{filepath} Warning: static and class properties don't need return type suffix")

    for change in changes:
        file_write(fix_file, change)

    global_changes = dict_update(global_changes, all_changes)

    return new_data


def check_camel_case(verif_file, fix_file, filepath, file_data):
    """Search and change wrong camel case"""

    global global_changes

    i = 0
    inside_class = None
    all_changes = {}

    nesting = 0

    new_data = ""

    while i < len(file_data):
        found_to_change = False
        variable = None
        lcc_variable = None

        if check_word_with_space(file_data[i:], "class") and not check_word_with_space(file_data[i:], "class var"):
            inside_class = read_next_word(file_data[i+6:])
            ucc_class = upper_camel_case(inside_class)

            new_data += file_data[i:i+6]
            i += 6

            if inside_class != ucc_class:
                class_changes = all_changes.get(inside_class)
                need_assign = False

                if class_changes == None:
                    class_changes = {}
                    all_changes = dict_update(
                        all_changes, {None: class_changes})
                    need_assign = True

                class_changes = dict_update(
                    class_changes, {inside_class: ucc_class})

                if need_assign:
                    all_changes = dict_update(
                        all_changes, {None: class_changes})

                new_data += ucc_class
                i += len(ucc_class)

        if check_word_with_space(file_data[i:], "var") or check_word_with_space(file_data[i:], "let"):
            variable = read_next_word(file_data[i+3:])
            lcc_variable = lower_camel_case(variable)

            new_data += file_data[i:i+4]
            i += 4

            found_to_change = True

        if check_word_with_space(file_data[i:], "func"):
            variable = read_next_word(file_data[i+4:])
            lcc_variable = lower_camel_case(variable)

            new_data += file_data[i:i+5]
            i += 5

            found_to_change = True

        if found_to_change:
            if variable != lcc_variable:
                class_changes = all_changes.get(inside_class)
                need_assign = False

                if class_changes == None:
                    class_changes = {}
                    all_changes = dict_update(
                        all_changes, {None: class_changes})
                    need_assign = True

                if class_changes.get(variable):
                    lcc_variable = lower_camel_case(
                        class_changes.get(variable))

                new_data += lcc_variable
                i += len(lcc_variable)

                class_changes = dict_update(
                    class_changes, {variable: lcc_variable})

                if need_assign:
                    all_changes = dict_update(
                        all_changes, {None: class_changes})

            found_to_change = False

        if file_data[i] == "{":
            nesting += 1

        if file_data[i] == "}":
            nesting -= 1

            if nesting == 0:  # class ended

                inside_class = None

        new_data += file_data[i]
        i += 1

    for class_changes in all_changes:
        for change_k in all_changes[class_changes]:
            file_write(
                fix_file, f"{filepath} Changed: wrong camel case on {change_k}")

    if len(all_changes) > 0:
        file_write(
            verif_file, f"{filepath} Error: wrong camel case declarations")

    global_changes = dict_update(global_changes, all_changes)

    return new_data


naming_fixers = [
    check_initializers,
    check_properties,
    check_camel_case
]
